@@ -1,7 +1,7 @@
 /*============================================================================
 
-  WCSLIB 4.17 - an implementation of the FITS WCS standard.
-  Copyright (C) 1995-2013, Mark Calabretta
+  WCSLIB 5.17 - an implementation of the FITS WCS standard.
+  Copyright (C) 1995-2017, Mark Calabretta
 
   This file is part of WCSLIB.
 
@@ -22,29 +22,32 @@
 
   Author: Mark Calabretta, Australia Telescope National Facility, CSIRO.
   http://www.atnf.csiro.au/people/Mark.Calabretta
-  $Id: lin.h,v 4.17 2013/01/29 05:29:20 cal103 Exp $
+  $Id: lin.h,v 5.17 2017/09/18 08:44:23 mcalabre Exp $
 *=============================================================================
 *
-* WCSLIB 4.17 - C routines that implement the FITS World Coordinate System
-* (WCS) standard.  Refer to
-*
-*   "Representations of world coordinates in FITS",
-*   Greisen, E.W., & Calabretta, M.R. 2002, A&A, 395, 1061 (Paper I)
-*
-* Refer to the README file provided with WCSLIB for an overview of the
-* library.
+* WCSLIB 5.17 - C routines that implement the FITS World Coordinate System
+* (WCS) standard.  Refer to the README file provided with WCSLIB for an
+* overview of the library.
 *
 *
 * Summary of the lin routines
 * ---------------------------
-* These routines apply the linear transformation defined by the FITS WCS
-* standard.  They are based on the linprm struct which contains all
-* information needed for the computations.  The struct contains some members
-* that must be set by the user, and others that are maintained by these
-* routines, somewhat like a C++ class but with no encapsulation.
+* Routines in this suite apply the linear transformation defined by the FITS
+* World Coordinate System (WCS) standard, as described in
 *
-* Three routines, linini(), lincpy(), and linfree() are provided to manage the
-* linprm struct, and another, linprt(), prints its contents.
+=   "Representations of world coordinates in FITS",
+=   Greisen, E.W., & Calabretta, M.R. 2002, A&A, 395, 1061 (WCS Paper I)
+*
+* These routines are based on the linprm struct which contains all information
+* needed for the computations.  The struct contains some members that must be
+* set by the user, and others that are maintained by these routines, somewhat
+* like a C++ class but with no encapsulation.
+*
+* Four routines, linini(), lindis(), lincpy(), and linfree() are provided to
+* manage the linprm struct, and another, linprt(), prints its contents.
+*
+* linperr() prints the error message(s) (if any) stored in a linprm struct,
+* and the disprm structs that it may contain.
 *
 * A setup routine, linset(), computes intermediate values in the linprm struct
 * from parameters in it that were supplied by the user.  The struct always
@@ -53,6 +56,9 @@
 *
 * linp2x() and linx2p() implement the WCS linear transformations.
 *
+* An auxiliary routine, linwarp(), computes various measures of the distortion
+* over a specified range of pixel coordinates.
+*
 * An auxiliary matrix inversion routine, matinv(), is included.  It uses
 * LU-triangular factorization with scaled partial pivoting.
 *
@@ -62,7 +68,7 @@
 * linini() allocates memory for arrays in a linprm struct and sets all members
 * of the struct to default values.
 *
-* PLEASE NOTE: every linprm struct should be initialized by linini(), possibly
+* PLEASE NOTE: every linprm struct must be initialized by linini(), possibly
 * repeatedly.  On the first invokation, and only the first invokation,
 * linprm::flag must be set to -1 to initialize memory management, regardless
 * of whether linini() will actually be used to allocate memory.
@@ -98,6 +104,43 @@
 *                       linprm::err if enabled, see wcserr_enable().
 *
 *
+* lindis() - Assign a distortion to a linprm struct
+* -------------------------------------------------
+* lindis() may be used to assign the address of a disprm struct to
+* linprm::dispre or linprm::disseq.  The linprm struct must already have been
+* initialized by linini().
+*
+* The disprm struct must have been allocated from the heap (e.g. using
+* malloc(), calloc(), etc.).  lindis() will immediately initialize it via a
+* call to disini() using the value of linprm::naxis.  Subsequently, it will be
+* reinitialized by calls to linini(), and freed by linfree(), neither of which
+* would happen if the disprm struct was assigned directly.
+*
+* If the disprm struct had previously been assigned via lindis(), it will be
+* freed before reassignment.  It is also permissable for a null disprm pointer
+* to be assigned to disable the distortion correction.
+*
+* Given:
+*   sequence  int       Is it a prior or sequent distortion?
+*                         1: Prior,   the assignment is to linprm::dispre.
+*                         2: Sequent, the assignment is to linprm::disseq.
+*
+*                       Anything else is an error.
+*
+* Given and returned:
+*   lin       struct linprm*
+*                       Linear transformation parameters.
+*
+*   dis       struct disprm*
+*                       Distortion function parameters.
+*
+* Function return value:
+*             int       Status return value:
+*                         0: Success.
+*                         1: Null linprm pointer passed.
+*                         4: Invalid sequence.
+*
+*
 * lincpy() - Copy routine for the linprm struct
 * ---------------------------------------------
 * lincpy() does a deep copy of one linprm struct to another, using linini() to
@@ -165,6 +208,26 @@
 *                         1: Null linprm pointer passed.
 *
 *
+* linperr() - Print error messages from a linprm struct
+* -----------------------------------------------------
+* linperr() prints the error message(s) (if any) stored in a linprm struct,
+* and the disprm structs that it may contain.  If there are no errors then
+* nothing is printed.  It uses wcserr_prt(), q.v.
+*
+* Given:
+*   lin       const struct linprm*
+*                       Coordinate transformation parameters.
+*
+*   prefix    const char *
+*                       If non-NULL, each output line will be prefixed with
+*                       this string.
+*
+* Function return value:
+*             int       Status return value:
+*                         0: Success.
+*                         1: Null linprm pointer passed.
+*
+*
 * linset() - Setup routine for the linprm struct
 * ----------------------------------------------
 * linset(), if necessary, allocates memory for the linprm::piximg and
@@ -251,6 +314,97 @@
 *                       linprm::err if enabled, see wcserr_enable().
 *
 *
+* linwarp() - Compute measures of distortion
+* ------------------------------------------
+* linwarp() computes various measures of the distortion over a specified range
+* of pixel coordinates.
+*
+* All distortion measures are specified as an offset in pixel coordinates,
+* as given directly by prior distortions.  The offset in intermediate pixel
+* coordinates given by sequent distortions is translated back to pixel
+* coordinates by applying the inverse of the linear transformation matrix
+* (PCi_ja or CDi_ja).  The difference may be significant if the matrix
+* introduced a scaling.
+*
+* If all distortions are prior, then linwarp() uses diswarp(), q.v.
+*
+* Given and returned:
+*   lin       struct linprm*
+*                       Linear transformation parameters plus distortions.
+*
+* Given:
+*   pixblc    const double[naxis]
+*                       Start of the range of pixel coordinates (i.e. "bottom
+*                       left-hand corner" in the conventional FITS image
+*                       display orientation).  May be specified as a NULL
+*                       pointer which is interpreted as (1,1,...).
+*
+*   pixtrc    const double[naxis]
+*                       End of the range of pixel coordinates (i.e. "top
+*                       right-hand corner" in the conventional FITS image
+*                       display orientation).
+*
+*   pixsamp   const double[naxis]
+*                       If positive or zero, the increment on the particular
+*                       axis, starting at pixblc[].  Zero is interpreted as a
+*                       unit increment.  pixsamp may also be specified as a
+*                       NULL pointer which is interpreted as all zeroes, i.e.
+*                       unit increments on all axes.
+*
+*                       If negative, the grid size on the particular axis (the
+*                       absolute value being rounded to the nearest integer).
+*                       For example, if pixsamp is (-128.0,-128.0,...) then
+*                       each axis will be sampled at 128 points between
+*                       pixblc[] and pixtrc[] inclusive.  Use caution when
+*                       using this option on non-square images.
+*
+* Returned:
+*   nsamp     int*      The number of pixel coordinates sampled.
+*
+*                       Can be specified as a NULL pointer if not required.
+*
+*   maxdis    double[naxis]
+*                       For each individual distortion function, the
+*                       maximum absolute value of the distortion.
+*
+*                       Can be specified as a NULL pointer if not required.
+*
+*   maxtot    double*   For the combination of all distortion functions, the
+*                       maximum absolute value of the distortion.
+*
+*                       Can be specified as a NULL pointer if not required.
+*
+*   avgdis    double[naxis]
+*                       For each individual distortion function, the
+*                       mean value of the distortion.
+*
+*                       Can be specified as a NULL pointer if not required.
+*
+*   avgtot    double*   For the combination of all distortion functions, the
+*                       mean value of the distortion.
+*
+*                       Can be specified as a NULL pointer if not required.
+*
+*   rmsdis    double[naxis]
+*                       For each individual distortion function, the
+*                       root mean square deviation of the distortion.
+*
+*                       Can be specified as a NULL pointer if not required.
+*
+*   rmstot    double*   For the combination of all distortion functions, the
+*                       root mean square deviation of the distortion.
+*
+*                       Can be specified as a NULL pointer if not required.
+*
+* Function return value:
+*             int       Status return value:
+*                         0: Success.
+*                         1: Null linprm pointer passed.
+*                         2: Memory allocation failed.
+*                         3: Invalid parameter.
+*                         4: Distort error.
+*
+*
 * linprm struct - Linear transformation parameters
 * ------------------------------------------------
 * The linprm struct contains all of the information required to perform a
@@ -264,7 +418,9 @@
 *
 *       - linprm::naxis (q.v., not normally set by the user),
 *       - linprm::pc,
-*       - linprm::cdelt.
+*       - linprm::cdelt,
+*       - linprm::dispre.
+*       - linprm::disseq.
 *
 *     This signals the initialization routine, linset(), to recompute the
 *     returned members of the linprm struct.  linset() will reset flag to
@@ -286,6 +442,9 @@
 *     (Given) Pointer to the first element of an array of double containing
 *     the coordinate reference pixel, CRPIXja.
 *
+*     It is not necessary to reset the linprm struct (via linset()) when
+*     linprm::crpix is changed.
+*
 *   double *pc
 *     (Given) Pointer to the first element of the PCi_ja (pixel coordinate)
 *     transformation matrix.  The expected order is
@@ -317,11 +476,56 @@
 *     (Given) Pointer to the first element of an array of double containing
 *     the coordinate increments, CDELTia.
 *
-*   int unity
-*     (Returned) True if the linear transformation matrix is unity.
+*   struct disprm *dispre
+*     (Given) Pointer to a disprm struct holding parameters for prior
+*     distortion functions, or a null (0x0) pointer if there are none.
 *
-*   int padding
-*     (An unused variable inserted for alignment purposes only.)
+*     Function lindis() may be used to assign a disprm pointer to a linprm
+*     struct, allowing it to take control of any memory allocated for it, as
+*     in the following example:
+*
+=       void add_distortion(struct linprm *lin)
+=       {
+=         struct disprm *dispre;
+=
+=         dispre = malloc(sizeof(struct disprm));
+=         dispre->flag = -1;
+=         lindis(1, lin, dispre);
+=           :
+=          (Set up dispre.)
+=           :
+=
+=         return;
+=       }
+*
+*     Here, after the distortion function parameters etc. are copied into
+*     dispre, dispre is assigned using lindis() which takes control of the
+*     allocated memory.  It will be free'd later when linfree() is invoked on
+*     the linprm struct.
+*
+*     Consider also the following erroneous code:
+*
+=       void bad_code(struct linprm *lin)
+=       {
+=         struct disprm dispre;
+=
+=         dispre.flag = -1;
+=         lindis(1, lin, &dispre);   // WRONG.
+=           :
+=
+=         return;
+=       }
+*
+*     Here, dispre is declared as a struct, rather than a pointer.  When the
+*     function returns, dispre will go out of scope and its memory will most
+*     likely be reused, thereby trashing its contents.  Later, a segfault will
+*     occur when linfree() tries to free dispre's stale address.
+*
+*   struct disprm *disseq
+*     (Given) Pointer to a disprm struct holding parameters for sequent
+*     distortion functions, or a null (0x0) pointer if there are none.
+*
+*     Refer to the comments and examples given for disprm::dispre.
 *
 *   double *piximg
 *     (Returned) Pointer to the first element of the matrix containing the
@@ -331,25 +535,38 @@
 *     (Returned) Pointer to the first element of the inverse of the
 *     linprm::piximg matrix.
 *
+*   int i_naxis
+*     (Returned) The dimension of linprm::piximg and linprm::imgpix (normally
+*     equal to naxis).
+*
+*   int unity
+*     (Returned) True if the linear transformation matrix is unity.
+*
+*   int affine
+*     (Returned) True if there are no distortions.
+*
+*   int simple
+*     (Returned) True if unity and no distortions.
+*
 *   struct wcserr *err
-*     (Returned) If enabled, when an error status is returned this struct
+*     (Returned) If enabled, when an error status is returned, this struct
 *     contains detailed information about the error, see wcserr_enable().
 *
-*   int i_naxis
+*   double *tmpcrd
 *     (For internal use only.)
 *   int m_flag
 *     (For internal use only.)
 *   int m_naxis
 *     (For internal use only.)
-*   int m_padding
-*     (For internal use only.)
 *   double *m_crpix
 *     (For internal use only.)
 *   double *m_pc
 *     (For internal use only.)
 *   double *m_cdelt
 *     (For internal use only.)
-*   void *padding2
+*   struct disprm *m_dispre
+*     (For internal use only.)
+*   struct disprm *m_disseq
 *     (For internal use only.)
 *
 *
@@ -362,8 +579,6 @@
 #ifndef WCSLIB_LIN
 #define WCSLIB_LIN
 
-#include "wcserr.h"
-
 #ifdef __cplusplus
 extern "C" {
 #endif
@@ -375,7 +590,10 @@ enum lin_errmsg_enum {
   LINERR_SUCCESS      = 0,	/* Success. */
   LINERR_NULL_POINTER = 1,	/* Null linprm pointer passed. */
   LINERR_MEMORY       = 2,	/* Memory allocation failed. */
-  LINERR_SINGULAR_MTX = 3 	/* PCi_ja matrix is singular. */
+  LINERR_SINGULAR_MTX = 3,	/* PCi_ja matrix is singular. */
+  LINERR_DISTORT_INIT = 4,	/* Failed to initialise distortions. */
+  LINERR_DISTORT      = 5,	/* Distort error. */
+  LINERR_DEDISTORT    = 6	/* De-distort error. */
 };
 
 struct linprm {
@@ -389,24 +607,29 @@ struct linprm {
   double *crpix;		/* CRPIXja keywords for each pixel axis.    */
   double *pc;			/* PCi_ja  linear transformation matrix.    */
   double *cdelt;		/* CDELTia keywords for each coord axis.    */
+  struct disprm *dispre;	/* Prior   distortion parameters, if any.   */
+  struct disprm *disseq;	/* Sequent distortion parameters, if any.   */
 
   /* Information derived from the parameters supplied.                      */
   /*------------------------------------------------------------------------*/
   double *piximg;		/* Product of CDELTia and PCi_ja matrices.  */
   double *imgpix;		/* Inverse of the piximg matrix.            */
+  int    i_naxis;		/* Dimension of piximg and imgpix.          */
   int    unity;			/* True if the PCi_ja matrix is unity.      */
+  int    affine;		/* True if there are no distortions.        */
+  int    simple;		/* True if unity and no distortions.        */
 
-  /* Error handling                                                         */
+  /* Error handling, if enabled.                                            */
   /*------------------------------------------------------------------------*/
-  int    padding;		/* (Dummy inserted for alignment purposes.) */
   struct wcserr *err;
 
-  /* Private - the remainder are for memory management.                     */
+  /* Private - the remainder are for internal use.                          */
   /*------------------------------------------------------------------------*/
-  int    i_naxis;
-  int    m_flag, m_naxis, m_padding;
+  double *tmpcrd;
+
+  int    m_flag, m_naxis;
   double *m_crpix, *m_pc, *m_cdelt;
-  void   *padding2;
+  struct disprm *m_dispre, *m_disseq;
 };
 
 /* Size of the linprm struct in int units, used by the Fortran wrappers. */
@@ -415,12 +638,16 @@ struct linprm {
 
 int linini(int alloc, int naxis, struct linprm *lin);
 
+int lindis(int sequence, struct linprm *lin, struct disprm *dis);
+
 int lincpy(int alloc, const struct linprm *linsrc, struct linprm *lindst);
 
 int linfree(struct linprm *lin);
 
 int linprt(const struct linprm *lin);
 
+int linperr(const struct linprm *lin, const char *prefix);
+
 int linset(struct linprm *lin);
 
 int linp2x(struct linprm *lin, int ncoord, int nelem, const double pixcrd[],
@@ -429,6 +656,12 @@ int linp2x(struct linprm *lin, int ncoord, int nelem, const double pixcrd[],
 int linx2p(struct linprm *lin, int ncoord, int nelem, const double imgcrd[],
            double pixcrd[]);
 
+int linwarp(struct linprm *lin, const double pixblc[], const double pixtrc[],
+            const double pixsamp[], int *nsamp,
+            double maxdis[], double *maxtot,
+            double avgdis[], double *avgtot,
+            double rmsdis[], double *rmstot);
+
 int matinv(int n, const double mat[], double inv[]);
 
 
